import { Test, TestingModule } from '@nestjs/testing';
import { HealthCheckError, HealthIndicatorResult } from '@nestjs/terminus';
import { RepositoriesHealthIndicatorUseCase } from '../../core/application/usecases/repositories.health-indicator.usecase';
import { MESSAGE_PUBLISHER } from '@src/app.di-tokens';
import { DatabaseErrorException } from '@mobicoop/ddd-library';
import {
  AUTHENTICATION_REPOSITORY,
  USERNAME_REPOSITORY,
} from '@modules/health/health.di-tokens';

const mockAuthenticationRepository = {
  healthCheck: jest
    .fn()
    .mockImplementationOnce(() => {
      return Promise.resolve(true);
    })
    .mockImplementation(() => {
      throw new DatabaseErrorException('An error occured in the database');
    }),
};

const mockUsernameRepository = {
  healthCheck: jest
    .fn()
    .mockImplementationOnce(() => {
      return Promise.resolve(true);
    })
    .mockImplementation(() => {
      throw new DatabaseErrorException('An error occured in the database');
    }),
};

const mockMessagePublisher = {
  publish: jest.fn().mockImplementation(),
};

describe('RepositoriesHealthIndicatorUseCase', () => {
  let repositoriesHealthIndicatorUseCase: RepositoriesHealthIndicatorUseCase;

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      providers: [
        RepositoriesHealthIndicatorUseCase,
        {
          provide: AUTHENTICATION_REPOSITORY,
          useValue: mockAuthenticationRepository,
        },
        {
          provide: USERNAME_REPOSITORY,
          useValue: mockUsernameRepository,
        },
        {
          provide: MESSAGE_PUBLISHER,
          useValue: mockMessagePublisher,
        },
      ],
    }).compile();

    repositoriesHealthIndicatorUseCase =
      module.get<RepositoriesHealthIndicatorUseCase>(
        RepositoriesHealthIndicatorUseCase,
      );
  });

  it('should be defined', () => {
    expect(repositoriesHealthIndicatorUseCase).toBeDefined();
  });

  describe('execute', () => {
    it('should check health successfully', async () => {
      const healthIndicatorResult: HealthIndicatorResult =
        await repositoriesHealthIndicatorUseCase.isHealthy('repositories');
      expect(healthIndicatorResult['repositories'].status).toBe('up');
    });

    it('should throw an error if database is unavailable', async () => {
      jest.spyOn(mockMessagePublisher, 'publish');
      await expect(
        repositoriesHealthIndicatorUseCase.isHealthy('repositories'),
      ).rejects.toBeInstanceOf(HealthCheckError);
      expect(mockMessagePublisher.publish).toHaveBeenCalledTimes(1);
    });
  });
});
